import argparse
from .vlab import Vlab
from os import getcwd
from . import __version__


class _VersionAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="show program's version number and exit"):
        super(_VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = self.version
        if version is None:
            version = parser.version
        formatter = parser._get_formatter()
        formatter.add_text(version)
        Vlab.check_sdk_version(False)
        parser.exit(message=formatter.format_help())


class JemuArgsParser:
    def __init__(self):
        self.args_parsed = None

    def parse(self, args):
        parser = argparse.ArgumentParser(
            prog='jumper',
            description="CLI interface for using Jumper's emulator"
        )

        parser.add_argument('--version', action=_VersionAction, version='%(prog)s {}'.format(__version__))
        parser.add_argument('--token', default=None, dest='token', type=str, help='Your secret token')

        subparsers = parser.add_subparsers(title='Commands', dest='command')

        run_parser = subparsers.add_parser(
            'run',
            help='Executes a FW file on a virtual device. Currently only support nRF52 devices'
        )
        run_parser.add_argument(
            '--fw',
            '--bin',
            '-b ',
            help="Firmware to be flashed to the virtual device (supported extensions are bin, out, elf, hex). In case more than one file needs to be flashed (such as using Nordic's softdevice), the files should be merged first. Check out https://vlab.jumper.io/docs#softdevice for more details",
        )

        run_parser.add_argument(
            '--debug-peripheral',
            action='store_true',
            help="Debug peripherals, enables to attach debugger to pid",
            default=False,
            dest='debug_peripheral'
        )

        run_parser.add_argument(
            '--directory',
            '-d ',
            help='Working directory, should include the board.json and scenario.json files. Default is current working directory',
            default=getcwd(),
            dest='working_directory'
        )

        run_parser.add_argument(
            '--sudo',
            '-s ',
            help='Run in sudo mode => FW can write to read-only registers. This should usually be used for testing low-level drivers, fuzzing (error injection) and certification tests.',
            action='store_true',
            default=False,
            dest='sudo_mode'
        )

        run_parser.add_argument(
            '--gdb',
            '-g ',
            help='Opens a GDB port for debugging the FW on port 5555. The FW will not start running until the GDB client connects.',
            action='store_true',
            default=False,
            dest='gdb_mode'
        )

        run_parser.add_argument(
            '--version',
            '-v ',
            help='Jumper sdk version.',
            action='store_true',
            default=False
        )

        run_parser.add_argument(
            '--trace',
            '-t ',
            help=
            """
            Prints a trace report to stdout.
            Valid reports: regs,interrupts,functions. (the functions trace can only be used with an out/elf file) 
            Example: jumper run --fw my_bin.bin -t interrupts,regs --trace-dest trace.txt
            Default value: regs 
            This can be used with --trace-dest to forward it to a file.
            """,
            const='regs',  # default when there are 0 arguments
            nargs='?',  # 0-or-1 arguments
            dest='traces_list'
        )

        run_parser.add_argument(
            '--trace-dest',
            type=str,
            help=
            """
            Forwards the trace report to a destination file. Must be used with -t.
            To print to stdout, just hit -t.
            """,
            default='',
            dest='trace_output_file'
        )

        run_parser.add_argument(
            '--uart',
            '-u ',
            action='store_true',
            default=False,
            help='Forwards UART prints to stdout, this can be used with --uart-dest to forward it to a file.',
            dest='print_uart'
        )

        run_parser.add_argument(
            '--uart-dest',
            type=str,
            help=
            """
            Forwards UART prints to a destination file. This MUST be used -u with this flag to make it work.
            To print to stdout, just hit -u.
            """,
            default=None,
            dest='uart_output_file'
        )

        run_parser.add_argument(
            '--uart-device',
            action='store_true',
            default=False,
            help='Creates a "uart" file in the working directory which is linked to the virtual UART device. Can be used with "screen". Linux only',
            dest='uart_device'
        )

        run_parser.add_argument(
            "--gpio",
            help=
            """
            Prints GPIO events (pin changes) to stdout.
            """,
            action='store_true',
            default=False,
        )

        run_parser.add_argument(
            '--platform',
            '-p ',
            type=str,
            choices=['nrf52832', 'stm32f4'],
            help=
            """
            Sets platform type. (Valid platforms: nrf52832,stm32f4).
            """,
            default=None,
            dest='platform'
        )

        run_parser.add_argument(
            '--stop-after',
            type=str,
            help=
            """
            Stop the execution after a specific amount of time. Units should be stated, if no units are stated, time in ms is assumed.
            Examples: "--stop-after 1s, --stop-after 1000ms, --stop-after 1000000us, --stop-after 1000000000ns" 
            """,
            default=None,
            dest='stop_after'
        )

        subparsers.add_parser(
            'ble',
            help='Creates a virtual HCI device (BLE dongle) for regular Linux/Bluez programs to communicate with virtual devices'
        )

        self.args_parsed = parser.parse_args(args)

    def get_args(self):
        return self.args_parsed


"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
import sys

from .__version__ import __version__
from .vlab import Vlab
from .vlab_hci_device import VirtualHciDevice, BluezError
from .jemu_interrupts import Interrupts
from .timeout_dec import timeout


class PythonVersionException(Exception):
    pass


if not sys.version_info[0] == 2 or not sys.version_info[1] == 7:
    raise PythonVersionException("python version should be 2.7")



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""

from time import sleep
import logging
import sys
from jemu_mem_peripheral import JemuMemPeripheral
import threading
from common import JemuConnectionException


class JemuSudo(JemuMemPeripheral):
    _id = None
    _jemu_connection = None
    _peripheral_type = None
    _state = None

    _DESCRIPTION = "description"
    _STOP_AFTER_COMMAND = "stop_after"
    _START_COMMAND = "resume_running"
    _COMMAND = "command"
    _NANO_SECONDS = "nanoseconds"
    _TYPE_STRING = "type"
    _PERIPHERAL_ID = "peripheral_id"
    _PERIPHERAL_TYPE = "peripheral_type"
    _STOPPED = "stopped"
    _RESUMED = "resumed"
    _COMMAND_SET_TIMER = "set_timer"
    _MESSAGE_ID = "message_id"
    _GET_STATE_COMMAND = "get_state"
    _CANCEL_STOP_ON_TICK = "cancel_stop"
    _CANCEL_STOP_RESPONSE = "stop_canceled"
    _GET_DEVICE_TIME = "get_device_time"
    _VALUE = "value"
    _TIMER_ID = "timer_id"
    _EMULATOR_STATE = "emulator_state"
    _DEVICE_TIME = "device_time"
    _SUDO_EXIT = "sudo_exit"
    _TERMINATE = "terminate"
    _PAUSE_COMMAND = "pause"

    _LOG_LEVEL = logging.ERROR

    def __init__(self, jemu_connection, id, peripheral_type):
        JemuMemPeripheral.__init__(self, jemu_connection, id, peripheral_type, None)
        self._id = id
        self._timer_id_counter = 0
        self._peripheral_type = peripheral_type
        self._jemu_connection = jemu_connection
        self._jemu_connection.register(self._receive_packet)
        self._timer_id_callback_dict = {}
        self._device_time_lock = threading.RLock()
        self._device_time = None
        self._stopped_packet_received = threading.Event()
        self._exit_code = None
        self._logger = logging.getLogger('JemuSudo')
        log_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        log_handler.setFormatter(formatter)
        self._logger.addHandler(log_handler)
        self._logger.setLevel(self._LOG_LEVEL)
        log_handler.setLevel(self._LOG_LEVEL)

    def _generate_json_command(self, command):
        return {
            self._COMMAND: command,
        }

    def stop_after_ns(self, nanoseconds):
        json_command = self._generate_json_command(self._STOP_AFTER_COMMAND)
        json_command[self._NANO_SECONDS] = nanoseconds
        self._jemu_connection.send_command_async(json_command, self._id, self._peripheral_type)

    def run_for_ns(self, nanoseconds):
        if self.get_state() == 'running':
            self._stopped_packet_received.clear()
            self.stop_after_ns(0)
            self._wait_for_event_from_connection(self._stopped_packet_received)

        if nanoseconds > 0:
            self._stopped_packet_received.clear()
            self.stop_after_ns(nanoseconds)
            self.resume()
            self._wait_for_event_from_connection(self._stopped_packet_received)

    def pause(self):
        response = self._jemu_connection.send_command(
            self._generate_json_command(self._PAUSE_COMMAND), self._id, self._peripheral_type
        )
        if not (self._DESCRIPTION in response and
                response[self._DESCRIPTION] == self._EMULATOR_STATE and
                self._VALUE in response and
                response[self._VALUE] == 'paused'):
            raise JemuConnectionException("Error: Couldn't resume jemu")

    def resume(self):
        if self.get_state() == 'paused':
            jemu_packet = self._jemu_connection.send_command(
                self._generate_json_command(self._START_COMMAND), self._id, self._peripheral_type
            )
            if not (self._DESCRIPTION in jemu_packet and jemu_packet[self._DESCRIPTION] == self._RESUMED):
                raise JemuConnectionException("Error: Couldn't resume jemu")

    def wait_until_stopped(self):
        self._stopped_packet_received.clear()
        self._wait_for_event_from_connection(self._stopped_packet_received)

    def _wait_for_event_from_connection(self, event):
        while not event.is_set():
            if not self._jemu_connection.is_connected():
                raise JemuConnectionException("Error: The Emulator was closed unexpectedly.")
            event.wait(0.2)

    def cancel_stop(self):
        jemu_packet = self._jemu_connection.send_command(self._generate_json_command(self._CANCEL_STOP_ON_TICK),
                                                         self._id, self._peripheral_type)
        if not (self._DESCRIPTION in jemu_packet and jemu_packet[self._DESCRIPTION] == self._CANCEL_STOP_RESPONSE):
            raise JemuConnectionException("Error: Couldn't cancel stop")

    def _receive_packet(self, jemu_packet):
        if self._DESCRIPTION in jemu_packet:
            if jemu_packet[self._DESCRIPTION] == self._STOPPED:
                self._stopped_packet_received.set()
            elif jemu_packet[self._DESCRIPTION] == self._TIMER_ID:
                cur_id = jemu_packet[self._VALUE]
                for id_from_dict in self._timer_id_callback_dict:
                    if id_from_dict == int(cur_id):
                        self._timer_id_callback_dict[int(cur_id)]()
            elif jemu_packet[self._DESCRIPTION] == self._SUDO_EXIT:
                self._exit_code = int(jemu_packet[self._VALUE])

    def set_timer(self, ticks, callback):
        self._timer_id_callback_dict.update({self._timer_id_counter: callback})
        json_command = self._generate_json_command(self._COMMAND_SET_TIMER)
        json_command[self._NANO_SECONDS] = ticks
        json_command[self._MESSAGE_ID] = self._timer_id_counter
        self._timer_id_counter += 1
        self._jemu_connection.send_command_async(json_command, self._id, self._peripheral_type)

    def get_state(self):
        self._logger.debug('Get state')
        jemu_packet = self._jemu_connection.send_command(self._generate_json_command(self._GET_STATE_COMMAND), self._id,
                                                         self._peripheral_type)
        self._logger.debug('Get state response:')
        self._logger.debug(jemu_packet)
        if jemu_packet is not None and \
                self._DESCRIPTION in jemu_packet and \
                jemu_packet[self._DESCRIPTION] == self._EMULATOR_STATE and \
                self._VALUE in jemu_packet:
            return jemu_packet[self._VALUE]
        else:
            raise JemuConnectionException("Error: Couldn't get jemu state")

    def get_device_time_ns(self):
        jemu_packet = self._jemu_connection.send_command(self._generate_json_command(self._GET_DEVICE_TIME), self._id,
                                                         self._peripheral_type)
        if self._DESCRIPTION in jemu_packet and jemu_packet[
            self._DESCRIPTION] == self._DEVICE_TIME and self._VALUE in jemu_packet:
            return jemu_packet[self._VALUE]
        else:
            raise JemuConnectionException("Error: Couldn't get device time")

    def get_exit_code(self):
        return self._exit_code

    def terminate(self):
        self._jemu_connection.send_command_async(
            self._generate_json_command(self._TERMINATE), self._id, self._peripheral_type
        )
